/*===- TableGen'erated file -------------------------------------*- C++ -*-===*\
|*                                                                            *|
|* Option Parsing Definitions                                                 *|
|*                                                                            *|
|* Automatically generated file, do not edit!                                 *|
|*                                                                            *|
\*===----------------------------------------------------------------------===*/

/////////
// String table

#ifdef OPTTABLE_STR_TABLE_CODE

#ifdef __GNUC__
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Woverlength-strings"
#endif
static constexpr char OptionStrTableStorage[] =
  "\0"
  "-\0"
  "--\0"
  "-?\0"
  "-??\0"
  "/\0"
  "/?\0"
  "/??\0"
  "<input>\0"
  "<unknown>\0"
  "/align:\0"
  "/aligncomm:\0"
  "/allowbind:no\0"
  "/allowbind\0"
  "/allowisolation:no\0"
  "/allowisolation\0"
  "/alternatename:\0"
  "/appcontainer:no\0"
  "/appcontainer\0"
  "/arm64xsameaddress:\0"
  "/assemblydebug:\0"
  "/assemblydebug\0"
  "/auto-import:no\0"
  "/auto-import\0"
  "/base:\0"
  "/Brepro\0"
  "/build-id:no\0"
  "/build-id\0"
  "/call-graph-ordering-file:\0"
  "/call-graph-profile-sort:no\0"
  "/call-graph-profile-sort\0"
  "/cetcompat:no\0"
  "/cetcompatstrict:no\0"
  "/cetcompatstrict\0"
  "/cetcompat\0"
  "/cetdynamicapisinproc:no\0"
  "/cetdynamicapisinproc\0"
  "/cetipvalidationrelaxed:no\0"
  "/cetipvalidationrelaxed\0"
  "--color-diagnostics=\0"
  "--color-diagnostics\0"
  "/debug:\0"
  "/debugtype:\0"
  "/debug\0"
  "/def:\0"
  "/defarm64native:\0"
  "/defaultlib:\0"
  "/delay:\0"
  "/delayload:\0"
  "/demangle:no\0"
  "/demangle\0"
  "/dependentloadflag:\0"
  "/dependentloadflag\0"
  "/diasdkdir:\0"
  "/disallowlib:\0"
  "/dll\0"
  "/driver:uponly,wdm\0"
  "/driver:uponly\0"
  "/driver:wdm,uponly\0"
  "/driver:wdm\0"
  "/driver\0"
  "/dwodir:\0"
  "/dynamicbase:no\0"
  "/dynamicbase\0"
  "/editandcontinue\0"
  "/emitpogophaseinfo\0"
  "/emittoolversioninfo:no\0"
  "/emittoolversioninfo\0"
  "/end-lib\0"
  "/entry:\0"
  "/errorlimit:\0"
  "/errorreport:\0"
  "/exclude-all-symbols\0"
  "/exclude-symbols:\0"
  "/export-all-symbols\0"
  "/export:\0"
  "/failifmismatch:\0"
  "/fastfail\0"
  "/fat-lto-objects:no\0"
  "/fat-lto-objects\0"
  "/filealign:\0"
  "/fixed:no\0"
  "/fixed\0"
  "/force:multipleres\0"
  "/force:multiple\0"
  "/force:unresolved\0"
  "/force\0"
  "/functionpadmin:\0"
  "/functionpadmin\0"
  "/guard:\0"
  "/guardsym:\0"
  "/heap:\0"
  "/help\0"
  "/highentropyva:no\0"
  "/highentropyva\0"
  "/hotpatchcompatible:no\0"
  "/hotpatchcompatible\0"
  "/idlout:\0"
  "/ignore:\0"
  "/ignoreidl\0"
  "/ilk:\0"
  "/implib:\0"
  "/include:\0"
  "/includeglob:\0"
  "/includeoptional:\0"
  "/incremental:no\0"
  "/incremental\0"
  "/inferasanlibs:no\0"
  "/inferasanlibs\0"
  "/integritycheck:no\0"
  "/integritycheck\0"
  "/kernel\0"
  "/kill-at\0"
  "/largeaddressaware:no\0"
  "/largeaddressaware\0"
  "/libpath:\0"
  "/lib\0"
  "/linkrepro:\0"
  "/linkreprofullpathrsp:\0"
  "/link\0"
  "/lld-allow-duplicate-weak:no\0"
  "/lld-allow-duplicate-weak\0"
  "/lldemit:\0"
  "/lldignoreenv\0"
  "/lldltocache:\0"
  "/lldltocachepolicy:\0"
  "/lldmap:\0"
  "/lldmap\0"
  "/lldmingw\0"
  "/lldsavetemps:\0"
  "/lldsavetemps\0"
  "/ltcg:\0"
  "/ltcgout:\0"
  "/ltcg\0"
  "/lto-cs-profile-file:\0"
  "/lto-cs-profile-generate\0"
  "/lto-obj-path:\0"
  "/lto-pgo-warn-mismatch:no\0"
  "/lto-pgo-warn-mismatch\0"
  "/lto-sample-profile:\0"
  "/machine:\0"
  "/manifest:\0"
  "/manifestdependency:\0"
  "/manifestfile:\0"
  "/manifestinput:\0"
  "/manifestuac:\0"
  "/manifest\0"
  "/map:\0"
  "/mapinfo:\0"
  "/map\0"
  "/maxilksize:\0"
  "/merge:\0"
  "/mllvm:\0"
  "/natvis:\0"
  "--no-color-diagnostics\0"
  "/nocoffgrpinfo\0"
  "/nodbgdirmerge\0"
  "/nodefaultlib:\0"
  "/nodefaultlib\0"
  "/noentry\0"
  "/noexp\0"
  "/noimplib\0"
  "/nologo\0"
  "/noseh\0"
  "/novcfeature\0"
  "/nxcompat:no\0"
  "/nxcompat\0"
  "/opt:\0"
  "/order:\0"
  "/osversion:\0"
  "/out:\0"
  "/output-def:\0"
  "/pdb:\0"
  "/pdbaltpath:\0"
  "/pdbcompress\0"
  "/pdbpagesize:\0"
  "/pdbsourcepath:\0"
  "/pdbstream:\0"
  "/pdbstripped:\0"
  "/prefetch-inputs:no\0"
  "/prefetch-inputs\0"
  "/print-search-paths\0"
  "/print-symbol-order:\0"
  "/profile\0"
  "/release\0"
  "/reproduce:\0"
  "--rsp-quoting=\0"
  "/runtime-pseudo-reloc:no\0"
  "/runtime-pseudo-reloc\0"
  "/safeseh:no\0"
  "/safeseh\0"
  "/section:\0"
  "/sectionlayout:\0"
  "/stack:\0"
  "/start-lib\0"
  "/stdcall-fixup:no\0"
  "/stdcall-fixup\0"
  "/stub:\0"
  "/subsystem:\0"
  "/summary\0"
  "/swaprun:cd\0"
  "/swaprun:net\0"
  "/swaprun:\0"
  "/thinlto-distributor-arg:\0"
  "/thinlto-distributor:\0"
  "/thinlto-emit-imports-files\0"
  "/thinlto-index-only:\0"
  "/thinlto-index-only\0"
  "/thinlto-object-suffix-replace:\0"
  "/thinlto-prefix-replace:\0"
  "/thinlto-remote-compiler-arg:\0"
  "/thinlto-remote-compiler-prepend-arg:\0"
  "/thinlto-remote-compiler:\0"
  "/threads:\0"
  "/throwingnew\0"
  "--time-trace-granularity=\0"
  "--time-trace=\0"
  "--time-trace\0"
  "/timestamp:\0"
  "/time\0"
  "/tlbid:\0"
  "/tlbout:\0"
  "/tsaware:no\0"
  "/tsaware\0"
  "/vctoolsdir:\0"
  "/vctoolsversion:\0"
  "/verbose:\0"
  "/verbose\0"
  "/version:\0"
  "--version\0"
  "/vfsoverlay:\0"
  "/wholearchive:\0"
  "/wholearchive\0"
  "/winsdkdir:\0"
  "/winsdkversion:\0"
  "/winsysroot:\0"
  "/wrap:\0"
  "/WX:no\0"
  "/WX\0"
  ;
#ifdef __GNUC__
#pragma GCC diagnostic pop
#endif

static constexpr llvm::StringTable
OptionStrTable = OptionStrTableStorage;
#endif // OPTTABLE_STR_TABLE_CODE

/////////
// Prefixes

#ifdef OPTTABLE_PREFIXES_TABLE_CODE
static constexpr llvm::StringTable::Offset OptionPrefixesTable[] = {
  0 /* prefixes */,
  1 /* prefixes */, 3 /* '--' */,
  4 /* prefixes */, 13 /* '/' */, 1 /* '-' */, 15 /* '/?' */, 6 /* '-?' */,
  4 /* prefixes */, 18 /* '/??' */, 9 /* '-??' */, 15 /* '/?' */, 6 /* '-?' */
};
#endif // OPTTABLE_PREFIXES_TABLE_CODE

/////////// SubCommand IDs

#ifdef OPTTABLE_SUBCOMMAND_IDS_TABLE_CODE
static constexpr unsigned OptionSubCommandIDsTable[] = {
  0 /* subcommands */
};
#endif // OPTTABLE_SUBCOMMAND_IDS_TABLE_CODE

/////////
// Prefix Union

#ifdef OPTTABLE_PREFIXES_UNION_CODE
static constexpr llvm::StringTable::Offset OptionPrefixesUnion[] = {
  1 /* '-' */,   3 /* '--' */,   6 /* '-?' */,   9 /* '-??' */,   13 /* '/' */,   15 /* '/?' */,   18 /* '/??' */
};
#endif // OPTTABLE_PREFIXES_UNION_CODE

/////////
// ValuesCode

#ifdef OPTTABLE_VALUES_CODE
#endif
/////////
// Groups

#ifdef OPTION

//////////
// Options

OPTION(0, 22 /* <input> */, INPUT, Input, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(0, 30 /* <unknown> */, UNKNOWN, Unknown, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 40 /* /align: */, align, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Section alignment", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 48 /* /aligncomm: */, aligncomm, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Set common symbol alignment", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 60 /* /allowbind:no */, allowbind_no, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Disable DLL binding", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 74 /* /allowbind */, allowbind, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Enable DLL binding (default)", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 85 /* /allowisolation:no */, allowisolation_no, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Disable DLL isolation", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 104 /* /allowisolation */, allowisolation, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Enable DLL isolation (default)", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 120 /* /alternatename: */, alternatename, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Define weak alias", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 136 /* /appcontainer:no */, appcontainer_no, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Image can run outside an app container (default)", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 153 /* /appcontainer */, appcontainer, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Image can only be run in an app container", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 167 /* /arm64xsameaddress: */, arm64xsameaddress, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Generate a thunk for the symbol with the same address in both native and EC views on ARM64X.", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 187 /* /assemblydebug: */, assemblydebug_opt, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 203 /* /assemblydebug */, assemblydebug, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 218 /* /auto-import:no */, auto_import_no, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 234 /* /auto-import */, auto_import, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 247 /* /base: */, base, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Base address of the program", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 254 /* /Brepro */, repro, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Use a hash of the executable as the PE header timestamp", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 262 /* /build-id:no */, build_id_no, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Do not Generate build ID", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 275 /* /build-id */, build_id, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Generate build ID (always on when generating PDB)", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 285 /* /call-graph-ordering-file: */, call_graph_ordering_file, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Layout sections to optimize the given callgraph", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 312 /* /call-graph-profile-sort:no */, call_graph_profile_sort_no, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Do not reorder sections with call graph profile", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 340 /* /call-graph-profile-sort */, call_graph_profile_sort, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Reorder sections with call graph profile (default)", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 365 /* /cetcompat:no */, cetcompat_no, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Don't mark executable image as compatible with Control-flow Enforcement Technology (CET) Shadow Stack (default)", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 379 /* /cetcompatstrict:no */, cetcompatstrict_no, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Don't mark executable image as compatible with Control-flow Enforcement Technology (CET) Shadow Stack in strict mode (default)", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 399 /* /cetcompatstrict */, cetcompatstrict, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Mark executable image as compatible with Control-flow Enforcement Technology (CET) Shadow Stack in strict mode", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 416 /* /cetcompat */, cetcompat, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Mark executable image as compatible with Control-flow Enforcement Technology (CET) Shadow Stack", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 427 /* /cetdynamicapisinproc:no */, cetdynamicapisinproc_no, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Don't mark executable image as compatible with Control-flow Enforcement Technology (CET) Shadow Stack with dynamic APIs allowed in process (default)", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 452 /* /cetdynamicapisinproc */, cetdynamicapisinproc, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Mark executable image as compatible with Control-flow Enforcement Technology (CET) Shadow Stack in such a way that dynamic APIs allowed in process", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 474 /* /cetipvalidationrelaxed:no */, cetipvalidationrelaxed_no, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Don't mark executable image as compatible with Control-flow Enforcement Technology (CET) Shadow Stack with relaxed context IP validation (default)", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 501 /* /cetipvalidationrelaxed */, cetipvalidationrelaxed, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Mark executable image as compatible with Control-flow Enforcement Technology (CET) Shadow Stack with relaxed context IP validation", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(1, 525 /* --color-diagnostics= */, color_diagnostics_eq, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Use colors in diagnostics (default: auto)", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), "[auto,always,never]", nullptr, 0)
OPTION(1, 546 /* --color-diagnostics */, color_diagnostics, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Alias for --color-diagnostics=always", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 566 /* /debug: */, debug_opt, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Embed a symbol table in the image with option", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 574 /* /debugtype: */, debugtype, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Debug Info Options", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 586 /* /debug */, debug, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Embed a symbol table in the image", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 593 /* /def: */, deffile, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Use module-definition file", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 599 /* /defarm64native: */, defarm64native, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Use a module-definition file for the native view in a hybrid image.", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 616 /* /defaultlib: */, defaultlib, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Add the library to the list of input files", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 629 /* /delay: */, delay, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 637 /* /delayload: */, delayload, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Delay loaded DLL name", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 649 /* /demangle:no */, demangle_no, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Do not demangle symbols in output", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 662 /* /demangle */, demangle, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Demangle symbols in output (default)", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 672 /* /dependentloadflag: */, dependentloadflag_opt, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Sets the default load flags used to resolve the statically linked imports of a module", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 692 /* /dependentloadflag */, dependentloadflag, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 711 /* /diasdkdir: */, diasdkdir, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Set the location of the DIA SDK", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 723 /* /disallowlib: */, disallowlib, Joined, INVALID, nodefaultlib, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 737 /* /dll */, dll, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Create a DLL", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 742 /* /driver:uponly,wdm */, driver_uponly_wdm, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 761 /* /driver:uponly */, driver_uponly, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Set IMAGE_DLL_CHARACTERISTICS_WDM_DRIVER bit in PE header", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 776 /* /driver:wdm,uponly */, driver_wdm_uponly, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 795 /* /driver:wdm */, driver_wdm, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Set IMAGE_FILE_UP_SYSTEM_ONLY bit in PE header", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 807 /* /driver */, driver, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Generate a Windows NT Kernel Mode Driver", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 815 /* /dwodir: */, dwodir, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Directory to store .dwo files when LTO and debug fission are used", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 824 /* /dynamicbase:no */, dynamicbase_no, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Disable ASLR (default when /fixed)", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 840 /* /dynamicbase */, dynamicbase, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Enable ASLR (default unless /fixed)", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 853 /* /editandcontinue */, editandcontinue, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 870 /* /emitpogophaseinfo */, emitpogophaseinfo, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 889 /* /emittoolversioninfo:no */, emittoolversioninfo_no, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Do not emit a tool version info after DOS header (so-called Rich header)", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 913 /* /emittoolversioninfo */, emittoolversioninfo, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Emit a tool version info after DOS header (so-called Rich header, default)", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 934 /* /end-lib */, end_lib, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "End group of objects treated as if they were in a library", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 943 /* /entry: */, entry, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Name of entry point symbol", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 951 /* /errorlimit: */, errorlimit, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Maximum number of errors to emit before stopping (0 = no limit)", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 964 /* /errorreport: */, errorreport, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 978 /* /exclude-all-symbols */, exclude_all_symbols, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 999 /* /exclude-symbols: */, exclude_symbols, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Exclude symbols from automatic export", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), "<symbol[,symbol,...]>", nullptr, 0)
OPTION(3, 1017 /* /export-all-symbols */, export_all_symbols, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1037 /* /export: */, export, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Export a function", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1046 /* /failifmismatch: */, failifmismatch, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1063 /* /fastfail */, fastfail, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1073 /* /fat-lto-objects:no */, fat_lto_objects_no, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Ignore the .llvm.lto section in relocatable object files (default).", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1093 /* /fat-lto-objects */, fat_lto_objects, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Use the .llvm.lto section, which contains LLVM bitcode, in fat LTO object files to perform LTO.", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1110 /* /filealign: */, filealign, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Section alignment in the output file", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1122 /* /fixed:no */, fixed_no, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Enable base relocations (default)", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1132 /* /fixed */, fixed, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Disable base relocations", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1139 /* /force:multipleres */, force_multipleres, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Allow multiply defined resources when creating executables", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1158 /* /force:multiple */, force_multiple, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Allow multiply defined symbols when creating executables", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1174 /* /force:unresolved */, force_unresolved, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Allow undefined symbols when creating executables", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1192 /* /force */, force, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Allow undefined and multiply defined symbols", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1199 /* /functionpadmin: */, functionpadmin_opt, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Prepares an image for hotpatching", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1216 /* /functionpadmin */, functionpadmin, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1232 /* /guard: */, guard, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Control flow guard", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1240 /* /guardsym: */, guardsym, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1251 /* /heap: */, heap, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Size of the heap", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1258 /* /help */, help, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1264 /* /highentropyva:no */, highentropyva_no, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Disable 64-bit ASLR", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1282 /* /highentropyva */, highentropyva, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Enable 64-bit ASLR (default on 64-bit)", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1297 /* /hotpatchcompatible:no */, hotpatchcompatible_no, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Don't mark executable image as compatible with hotpatch (default)", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1320 /* /hotpatchcompatible */, hotpatchcompatible, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Mark executable image as compatible with hotpatch", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1340 /* /idlout: */, idlout, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1349 /* /ignore: */, ignore, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Specify warning codes to ignore", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1358 /* /ignoreidl */, ignoreidl, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1369 /* /ilk: */, ilk, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1375 /* /implib: */, implib, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Import library name", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1384 /* /include: */, incl, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Force symbol to be added to symbol table as undefined one", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1394 /* /includeglob: */, incl_glob, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Force symbol to be added to symbol table as undefined one using a glob pattern", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1408 /* /includeoptional: */, include_optional, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Add symbol as undefined, but allow it to remain undefined", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1426 /* /incremental:no */, incremental_no, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Overwrite import library even if contents are unchanged", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1442 /* /incremental */, incremental, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Keep original import library if contents are unchanged", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1455 /* /inferasanlibs:no */, inferasanlibs_no, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "No effect (default)", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1473 /* /inferasanlibs */, inferasanlibs, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Unused, generates a warning", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1488 /* /integritycheck:no */, integritycheck_no, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "No effect (default)", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1507 /* /integritycheck */, integritycheck, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Set FORCE_INTEGRITY bit in PE header", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1523 /* /kernel */, kernel, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1531 /* /kill-at */, kill_at, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1540 /* /largeaddressaware:no */, largeaddressaware_no, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Disable large addresses (default on 32-bit)", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1562 /* /largeaddressaware */, largeaddressaware, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Enable large addresses (default on 64-bit)", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1581 /* /libpath: */, libpath, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Additional library search path", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1591 /* /lib */, lib, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Act like lib.exe; must be first argument if present", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1596 /* /linkrepro: */, linkrepro, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Write repro.tar containing inputs and command to reproduce link", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), "directory", nullptr, 0)
OPTION(3, 1608 /* /linkreprofullpathrsp: */, linkreprofullpathrsp, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Write .rsp file containing inputs used to link with full paths", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), "directory", nullptr, 0)
OPTION(3, 1631 /* /link */, link, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Ignored for compatibility", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1637 /* /lld-allow-duplicate-weak:no */, lld_allow_duplicate_weak_no, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1666 /* /lld-allow-duplicate-weak */, lld_allow_duplicate_weak, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1692 /* /lldemit: */, lldemit, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Specify output type", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1702 /* /lldignoreenv */, lldignoreenv, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Ignore environment variables like %LIB%", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1716 /* /lldltocache: */, lldltocache, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Path to ThinLTO cached object file directory", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1730 /* /lldltocachepolicy: */, lldltocachepolicy, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Pruning policy for the ThinLTO cache", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1750 /* /lldmap: */, lldmap_file, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1759 /* /lldmap */, lldmap, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1767 /* /lldmingw */, lldmingw, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1777 /* /lldsavetemps: */, lldsavetemps_colon, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Save select intermediate LTO compilation results", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, "resolution,preopt,promote,internalize,import,opt,precodegen,prelink,combinedindex", 0)
OPTION(3, 1792 /* /lldsavetemps */, lldsavetemps, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Save intermediate LTO compilation results", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1806 /* /ltcg: */, ltcg_opt, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1813 /* /ltcgout: */, ltcgout, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1823 /* /ltcg */, ltcg, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1829 /* /lto-cs-profile-file: */, lto_cs_profile_file, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Context sensitive profile file path", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1851 /* /lto-cs-profile-generate */, lto_cs_profile_generate, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Perform context sensitive PGO instrumentation", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1876 /* /lto-obj-path: */, lto_obj_path, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "output native object for merged LTO unit to this path", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1891 /* /lto-pgo-warn-mismatch:no */, lto_pgo_warn_mismatch_no, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "turn off warnings about profile cfg mismatch", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1917 /* /lto-pgo-warn-mismatch */, lto_pgo_warn_mismatch, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "turn on warnings about profile cfg mismatch (default)>", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1940 /* /lto-sample-profile: */, lto_sample_profile, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Sample profile file path", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1961 /* /machine: */, machine, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Specify target platform", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1971 /* /manifest: */, manifest_colon, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "NO disables manifest output; EMBED[,ID=#] embeds manifest as resource in the image", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 1982 /* /manifestdependency: */, manifestdependency, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Attributes for <dependency> element in manifest file; implies /manifest", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2003 /* /manifestfile: */, manifestfile, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Manifest output path, with /manifest", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2018 /* /manifestinput: */, manifestinput, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Additional manifest inputs; only valid with /manifest:embed", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2034 /* /manifestuac: */, manifestuac, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "User access control", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2048 /* /manifest */, manifest, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Create .manifest file", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2058 /* /map: */, map_file, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2064 /* /mapinfo: */, map_info, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Include the specified information in a map file", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2074 /* /map */, map, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2079 /* /maxilksize: */, maxilksize, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2092 /* /merge: */, merge, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Combine sections", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2100 /* /mllvm: */, mllvm, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Options to pass to LLVM", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2108 /* /natvis: */, natvis, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Path to natvis file to embed in the PDB", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(1, 2117 /* --no-color-diagnostics */, no_color_diagnostics, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Alias for --color-diagnostics=never", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2140 /* /nocoffgrpinfo */, nocoffgrpinfo, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2155 /* /nodbgdirmerge */, nodbgdirmerge, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Emit the debug directory in a separate section", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2170 /* /nodefaultlib: */, nodefaultlib, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Remove a default library", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2185 /* /nodefaultlib */, nodefaultlib_all, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Remove all default libraries", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2199 /* /noentry */, noentry, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Don't add reference to DllMainCRTStartup; only valid with /dll", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2208 /* /noexp */, noexp, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2215 /* /noimplib */, noimplib, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Don't output an import lib", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2225 /* /nologo */, nologo, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2233 /* /noseh */, noseh, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2240 /* /novcfeature */, novcfeature, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2253 /* /nxcompat:no */, nxcompat_no, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Disable data execution provention", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2266 /* /nxcompat */, nxcompat, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Enable data execution prevention (default)", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2276 /* /opt: */, opt, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Control optimizations", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2282 /* /order: */, order, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Put functions in order", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2290 /* /osversion: */, osversion, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2302 /* /out: */, out, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Path to file to write output", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2308 /* /output-def: */, output_def, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2321 /* /pdb: */, pdb, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "PDB file path", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2327 /* /pdbaltpath: */, pdbaltpath, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "PDB file path to embed in the image", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2340 /* /pdbcompress */, pdbcompress, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2353 /* /pdbpagesize: */, pdbpagesize, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "PDB page size", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2367 /* /pdbsourcepath: */, pdb_source_path, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Base path used to make relative source file path absolute in PDB", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2383 /* /pdbstream: */, pdbstream, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Embed the contents of <file> in the PDB as named stream <name>", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), "<name>=<file>", nullptr, 0)
OPTION(3, 2395 /* /pdbstripped: */, pdbstripped, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Stripped PDB file path", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2409 /* /prefetch-inputs:no */, prefetch_inputs_no, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Do not prefetch input files (default)", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2429 /* /prefetch-inputs */, prefetch_inputs, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Request the OS to prefetch input files as early as possible, to improve link times", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2446 /* /print-search-paths */, print_search_paths, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2466 /* /print-symbol-order: */, print_symbol_order, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Print a symbol order specified by /call-graph-ordering-file and /call-graph-profile-sort into the specified file", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2487 /* /profile */, profile, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2496 /* /release */, release, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Set the Checksum in the header of an PE file", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2505 /* /reproduce: */, reproduce, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Write tar file containing inputs and command to reproduce link", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), "filename", nullptr, 0)
OPTION(1, 2517 /* --rsp-quoting= */, rsp_quoting, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Quoting style for response files, 'windows' (default) or 'posix'", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2532 /* /runtime-pseudo-reloc:no */, runtime_pseudo_reloc_no, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2557 /* /runtime-pseudo-reloc */, runtime_pseudo_reloc, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2579 /* /safeseh:no */, safeseh_no, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Don't produce an image with Safe Exception Handler", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2591 /* /safeseh */, safeseh, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Produce an image with Safe Exception Handler (only for x86)", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2600 /* /section: */, section, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Specify section attributes", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2610 /* /sectionlayout: */, sectionlayout, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Specifies the layout strategy for output sections", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2626 /* /stack: */, stack, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Size of the stack", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2634 /* /start-lib */, start_lib, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Start group of objects treated as if they were in a library", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2645 /* /stdcall-fixup:no */, stdcall_fixup_no, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2663 /* /stdcall-fixup */, stdcall_fixup, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2678 /* /stub: */, stub, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Specify DOS stub file", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2685 /* /subsystem: */, subsystem, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Specify subsystem", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2697 /* /summary */, summary, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2706 /* /swaprun:cd */, swaprun_cd, Flag, INVALID, swaprun, "cd\0", 0, DefaultVis, 0,
       "Make loader run output binary from swap instead of from CD", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2718 /* /swaprun:net */, swaprun_net, Flag, INVALID, swaprun, "net\0", 0, DefaultVis, 0,
       "Make loader run output binary from swap instead of from network", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2731 /* /swaprun: */, swaprun, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Comma-separated list of 'cd' or 'net'", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2741 /* /thinlto-distributor-arg: */, thinlto_distributor_arg, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Arguments to pass to the ThinLTO distributor", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2767 /* /thinlto-distributor: */, thinlto_distributor, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Distributor to use for ThinLTO backend compilations. If specified, ThinLTO backend compilations will be distributed", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2789 /* /thinlto-emit-imports-files */, thinlto_emit_imports_files, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Emit .imports files with -thinlto-index-only", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2817 /* /thinlto-index-only: */, thinlto_index_only_arg, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "-thinlto-index-only and also write native module names to file", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2838 /* /thinlto-index-only */, thinlto_index_only, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Instead of linking, emit ThinLTO index files", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2858 /* /thinlto-object-suffix-replace: */, thinlto_object_suffix_replace, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "'old;new' replace old suffix with new suffix in ThinLTO index", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2890 /* /thinlto-prefix-replace: */, thinlto_prefix_replace, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "'old;new' replace old prefix with new prefix in ThinLTO outputs", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2915 /* /thinlto-remote-compiler-arg: */, thinlto_remote_compiler_arg, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Compiler arguments for the ThinLTO distributor to pass for ThinLTO backend compilations", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2945 /* /thinlto-remote-compiler-prepend-arg: */, thinlto_remote_compiler_prepend_arg, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Compiler prepend arguments for the ThinLTO distributor to pass for ThinLTO backend compilations", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 2983 /* /thinlto-remote-compiler: */, thinlto_remote_compiler, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Compiler for the ThinLTO distributor to invoke for ThinLTO backend compilations", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 3009 /* /threads: */, threads, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Number of threads. '1' disables multi-threading. By default all available hardware threads are used", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 3019 /* /throwingnew */, throwingnew, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(1, 3032 /* --time-trace-granularity= */, time_trace_granularity_eq, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Minimum time granularity (in microseconds) traced by time profiler", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(1, 3058 /* --time-trace= */, time_trace_eq, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Record time trace to <file>", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), "<file>", nullptr, 0)
OPTION(1, 3072 /* --time-trace */, anonymous_0, Flag, INVALID, time_trace_eq, nullptr, 0, DefaultVis, 0,
       "Record time trace to file next to output", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 3085 /* /timestamp: */, timestamp, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Specify the PE header timestamp", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 3097 /* /time */, show_timing, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Print the time spent in each phase of linking", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 3103 /* /tlbid: */, tlbid, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 3111 /* /tlbout: */, tlbout, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 3120 /* /tsaware:no */, tsaware_no, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Create non-Terminal Server aware executable", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 3132 /* /tsaware */, tsaware, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Create Terminal Server aware executable (default)", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 3141 /* /vctoolsdir: */, vctoolsdir, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Set the location of the VC tools", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 3154 /* /vctoolsversion: */, vctoolsversion, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Specify which VC tools version to use", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 3171 /* /verbose: */, verbose_all, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 3181 /* /verbose */, verbose, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 3190 /* /version: */, version, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Specify a version number in the PE header", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(1, 3200 /* --version */, dash_dash_version, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Display the version number and exit", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 3210 /* /vfsoverlay: */, vfsoverlay, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Path to a vfsoverlay yaml file to optionally look for /defaultlib's in", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 3223 /* /wholearchive: */, wholearchive_file, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Include all object files from this library", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 3238 /* /wholearchive */, wholearchive_flag, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Include all object files from all libraries", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 3252 /* /winsdkdir: */, winsdkdir, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Set the location of the Windows SDK", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 3264 /* /winsdkversion: */, winsdkversion, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Specify which SDK version to use", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 3280 /* /winsysroot: */, winsysroot, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Adds several subdirectories to the library search paths", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 3293 /* /wrap: */, wrap, Joined, INVALID, INVALID, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 3300 /* /WX:no */, WX_no, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Don't treat warnings as errors (default)", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(3, 3307 /* /WX */, WX, Flag, INVALID, INVALID, nullptr, 0, DefaultVis, 0,
       "Treat warnings as errors", (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
OPTION(8, 18 /* /?? */, help_q, Flag, INVALID, help, nullptr, 0, DefaultVis, 0, nullptr, (std::array<std::pair<std::array<unsigned, 2>, const char*>, 1>{{ {std::array<unsigned, 2>{{0, 0}}, nullptr} }}), nullptr, nullptr, 0)
#endif // OPTION

#ifdef SIMPLE_ENUM_VALUE_TABLE

struct SimpleEnumValue {
  const char *Name;
  unsigned Value;
};

struct SimpleEnumValueTable {
  const SimpleEnumValue *Table;
  unsigned Size;
};
static const SimpleEnumValueTable SimpleEnumValueTables[] = {};
static const unsigned SimpleEnumValueTablesSize = std::size(SimpleEnumValueTables);
#endif // SIMPLE_ENUM_VALUE_TABLE

/////////

// SubCommands

#ifdef OPTTABLE_SUBCOMMANDS_CODE
static constexpr llvm::opt::OptTable::SubCommand OptionSubCommands[] = {
};
#endif // OPTTABLE_SUBCOMMANDS_CODE


