#ifndef PCREL_OPT
#error "Need to define function-style macro PCREL_OPT"
#endif
PCREL_OPT(NOINSN, NOINSN, NOMASK);
PCREL_OPT(LBZ, PLBZ, OPC_AND_RST);
PCREL_OPT(LHZ, PLHZ, OPC_AND_RST);
PCREL_OPT(LWZ, PLWZ, OPC_AND_RST);
PCREL_OPT(LHA, PLHA, OPC_AND_RST);
PCREL_OPT(LWA, PLWA, ONLY_RST);
PCREL_OPT(LD, PLD , ONLY_RST);
PCREL_OPT(LFS, PLFS, OPC_AND_RST);
PCREL_OPT(LXSSP, PLXSSP, ONLY_RST);
PCREL_OPT(LFD, PLFD, OPC_AND_RST);
PCREL_OPT(LXSD, PLXSD, ONLY_RST);
PCREL_OPT(LXV, PLXV, ST_STX28_TO5);
PCREL_OPT(LXVP, PLXVP, OPC_AND_RST);

PCREL_OPT(STB, PSTB, OPC_AND_RST);
PCREL_OPT(STH, PSTH, OPC_AND_RST);
PCREL_OPT(STW, PSTW, OPC_AND_RST);
PCREL_OPT(STD, PSTD, ONLY_RST);
PCREL_OPT(STFS, PSTFS, OPC_AND_RST);
PCREL_OPT(STXSSP, PSTXSSP, ONLY_RST);
PCREL_OPT(STFD, PSTFD, OPC_AND_RST);
PCREL_OPT(STXSD, PSTXSD, ONLY_RST);
PCREL_OPT(STXV, PSTXV, ST_STX28_TO5);
PCREL_OPT(STXVP, PSTXVP, OPC_AND_RST);
